"""Messenger gRPC grpc-messenger-client."""
import warnings
from datetime import datetime
from typing import Dict, Generator, List, Optional, Type, TypeVar, Union
from uuid import UUID

import grpc
from fred_api.messenger import service_email_grpc_pb2_grpc, service_sms_grpc_pb2_grpc
from fred_api.messenger.common_types_pb2 import Reference, ReferenceType, Uuid
from fred_api.messenger.service_email_grpc_pb2 import GetEmailRequest, ListEmailRequest, SendEmailRequest
from fred_api.messenger.service_sms_grpc_pb2 import GetSmsRequest, ListSmsRequest, SendSmsRequest
from frgal import GrpcClient, GrpcDecoder
from google.protobuf.struct_pb2 import Struct
from typist import SecretaryClient

from .email import Email, EmailEnvelope
from .message import MessageEnvelope
from .sms import Sms, SmsEnvelope

E = TypeVar('E', bound=MessageEnvelope)


class HermesError(Exception):
    """Base class for Hermes related errors."""


class NoMessage(HermesError):
    """Raised when no message matching the UUID is found."""


class SendFailed(HermesError):
    """Raised when sending of the message failed."""


class MessengerDecoder(GrpcDecoder):
    """Messenger gRPC decoder."""

    def __init__(self):
        """Initialize the decoder object by registering custom decoders."""
        super().__init__()
        self.set_decoder(Reference, self._decode_reference)
        self.set_decoder(Struct, self._decode_protobuf_struct)
        self.set_decoder(Uuid, self._decode_uuid)

    def _decode_reference(self, value: Reference) -> Dict[str, Union[UUID, str]]:
        """Decode Reference."""
        if value.type == ReferenceType.registrar:
            return {'type': ReferenceType.Name(value.type), 'value': value.value}
        else:
            return {'type': ReferenceType.Name(value.type), 'value': UUID(value.value)}

    def _decode_protobuf_struct(self, value: Struct) -> Dict:
        """Decode google.protobuf.Struct to Dict."""
        return {k: self.decode(v) for k, v in value.items()}

    def _decode_uuid(self, value: Uuid) -> Optional[UUID]:
        """Decode Uuid."""
        if not value.value:
            return None
        return UUID(value.value)


class MessengerClient(GrpcClient):
    """Messenger gRPC grpc-messenger-client."""

    email_service = "EmailMessenger"
    sms_service = "SmsMessenger"

    def __init__(
        self,
        netloc: str,
        credentials: grpc.ChannelCredentials = None,
        *,
        secretary_client: SecretaryClient = None
    ):
        """Initialize Logger GRPC backend grpc-messenger-client instance.

        Args:
            netloc: Network location of a gRPC server.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
            secretary_client: Instance of django-secretary REST API client.

        """
        decoder = MessengerDecoder()
        super().__init__(
            netloc,
            decoder=decoder,
            grpc_modules=[service_email_grpc_pb2_grpc, service_sms_grpc_pb2_grpc],
            credentials=credentials,
        )
        self.secretary_client = secretary_client

    def send(self, email: Email, *, archive: bool = True) -> str:
        """Send gRPC request to send email request.

        Deprecated alias of send_email.
        """
        warnings.warn("Method send is deprecated. Use send_email instead.", DeprecationWarning)
        return self.send_email(email, archive=archive)

    def get(self, uid: str) -> EmailEnvelope:
        """Send gRPC request to get e-mail.

        Deprecated alias of get_email.
        """
        warnings.warn("Method get is deprecated. Use get_email instead.", DeprecationWarning)
        return self.get_email(uid)

    def list(
        self,
        created_from: datetime = None,
        created_to: datetime = None,
        recipients: List[str] = None,
        body_templates: List[str] = None,
        types: List[str] = None,
        limit: int = None
    ) -> Generator[EmailEnvelope, None, None]:
        """Send gRPC request to get e-mail list.

        Deprecated alias of list_emails.
        """
        warnings.warn("Method list is deprecated. Use list_emails instead.", DeprecationWarning)
        return self.list_emails(
            created_from=created_from,
            created_to=created_to,
            recipients=recipients,
            body_templates=body_templates,
            types=types,
            limit=limit,
        )

    def send_email(self, email: Email, *, archive: bool = True) -> str:
        """Send gRPC request to send email request.

        Args:
            email: Email structure to be sent
            archive: Whether the message should be archived
        Raises:
            SendFailed: When sending of the message fails.
        """
        request = SendEmailRequest()
        request.message_data.recipient = email.recipient
        request.message_data.subject_template = email.subject_template
        request.message_data.body_template = email.body_template
        request.message_data.extra_headers.update(**(email.extra_headers or {}))
        request.message_data.context.update(email.context)
        if email.type:
            request.message_data.type = email.type
        request.archive = archive
        if email.sender is not None:
            request.message_data.sender = email.sender
        if email.body_template_html is not None:
            request.message_data.body_template_html = email.body_template_html
        if email.subject_template_uuid is not None:
            request.message_data.subject_template_uuid.value = str(email.subject_template_uuid)
        if email.body_template_uuid is not None:
            request.message_data.body_template_uuid.value = str(email.body_template_uuid)
        if email.body_template_html_uuid is not None:
            request.message_data.body_template_html_uuid.value = str(email.body_template_html_uuid)
        for uuid in email.attachments or ():
            request.message_data.attachments.add(value=str(uuid))
        try:
            return self.call(self.email_service, "send", request)
        except grpc.RpcError as error:
            raise SendFailed('Sending of the email failed: {}'.format(error))

    def get_email(self, uid: str) -> EmailEnvelope:
        """Send gRPC request to get e-mail.

        Args:
            uid: Unique mail identification.

        Raises:
            NoMessage: When no message matching uid is found.
        """
        request = GetEmailRequest()
        request.uid.value = uid
        try:
            message_data = self.call(self.email_service, "get", request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise NoMessage('Message {} not found'.format(uid))
            raise
        return EmailEnvelope.from_dict(message_data, secretary_client=self.secretary_client)

    def list_emails(self, **kwargs) -> Generator[EmailEnvelope, None, None]:
        """Send gRPC request to get email list."""
        return self._list_messages(
            grpc_service=self.email_service,
            request_cls=ListEmailRequest,
            envelope_cls=EmailEnvelope,
            **kwargs,
        )

    def send_sms(self, sms: Sms, *, archive: bool = True) -> str:
        """Send gRPC request to send sms request.

        Args:
            sms: Sms structure to be sent
            archive: Whether the message should be archived
        Raises:
            SendFailed: When sending of the message fails.
        """
        request = SendSmsRequest()
        request.message_data.recipient = sms.recipient
        request.message_data.body_template = sms.body_template
        request.message_data.context.update(sms.context)
        request.archive = archive
        if sms.type:
            request.message_data.type = sms.type
        if sms.body_template_uuid is not None:
            request.message_data.body_template_uuid.value = str(sms.body_template_uuid)
        try:
            return self.call(self.sms_service, "send", request)
        except grpc.RpcError as error:
            raise SendFailed('Sending of the sms failed: {}'.format(error))

    def get_sms(self, uid: str) -> SmsEnvelope:
        """Send gRPC request to get sms.

        Args:
            uid: Unique mail identification.

        Raises:
            NoMessage: When no message matching uid is found.
        """
        request = GetSmsRequest()
        request.uid.value = uid
        try:
            message_data = self.call(self.sms_service, "get", request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise NoMessage('Message {} not found'.format(uid))
            raise
        return SmsEnvelope.from_dict(message_data, secretary_client=self.secretary_client)

    def list_sms(self, **kwargs) -> Generator[SmsEnvelope, None, None]:
        """Send gRPC request to get sms list."""
        return self._list_messages(
            grpc_service=self.sms_service,
            request_cls=ListSmsRequest,
            envelope_cls=SmsEnvelope,
            **kwargs,
        )

    def _list_messages(
        self,
        grpc_service: str,
        request_cls: Union[ListEmailRequest, ListSmsRequest],
        envelope_cls: Type[E],
        created_from: datetime = None,
        created_to: datetime = None,
        recipients: List[str] = None,
        body_templates: List[str] = None,
        types: List[str] = None,
        limit: int = None
    ) -> Generator[E, None, None]:
        """Send gRPC request to get list of messages.

        Args:
            grpc_service: Name of grpc service implementing ``list`` method.
            request_cls: List request message class.
            envelope_cls: Message envelope class.

            created_from: Lower limit for create datetime.
            created_to: Upper limit for create datetime.
            recipients: List of recipients to filter.
            body_templates: List of body template names to filter.
            types: List of message types to filter.
            limit: Maximum number of messages returned.
        """
        request = request_cls()
        if created_from:
            request.created_from.FromDatetime(created_from)
        if created_to:
            request.created_to.FromDatetime(created_to)
        if recipients:
            request.recipients[:] = recipients
        if body_templates:
            request.body_templates[:] = body_templates
        if types:
            request.types[:] = types
        if limit:
            request.limit = limit

        message_list = self.call_stream(grpc_service, "list", request)

        for message_data in message_list:
            yield envelope_cls.from_dict(message_data, secretary_client=self.secretary_client)
