"""Unittests for SMS structures."""
from unittest import TestCase
from unittest.mock import Mock, call, sentinel
from uuid import UUID

from typist import SecretaryClient

from hermes.sms import Sms


class SmsTest(TestCase):
    """Unittests for Sms structure."""

    def test_init_empty(self):
        sms = Sms(sentinel.recipient, sentinel.body_template)
        self.assertEqual(sms.recipient, sentinel.recipient)
        self.assertEqual(sms.body_template, sentinel.body_template)
        self.assertEqual(sms.context, {})

    def test_init_values(self):
        sms = Sms(
            sentinel.recipient, sentinel.body_template,
            context=sentinel.context,
            body_template_uuid=UUID(int=1),
        )
        self.assertEqual(sms.recipient, sentinel.recipient)
        self.assertEqual(sms.body_template, sentinel.body_template)
        self.assertEqual(sms.body_template_uuid, UUID(int=1))
        self.assertEqual(sms.context, sentinel.context)

    def test_from_dict_missing_recipient(self):
        with self.assertRaises(ValueError):
            Sms.from_dict({'body_template': sentinel.body_template})

    def test_from_dict_missing_body_template(self):
        with self.assertRaises(ValueError):
            Sms.from_dict({'recipient': sentinel.recipient})

    def test_from_dict_data(self):
        sms = Sms.from_dict({
            'recipient': sentinel.recipient,
            'body_template': sentinel.body_template,
            'context': {},
            'body_template_uuid': UUID(int=1),
            'type': sentinel.type,
        }, secretary_client=sentinel.secretary_client)

        self.assertEqual(sms.recipient, sentinel.recipient)
        self.assertEqual(sms.body_template, sentinel.body_template)
        self.assertEqual(sms.context, {})
        self.assertEqual(sms.body_template_uuid, UUID(int=1))
        self.assertEqual(sms.type, sentinel.type)
        self.assertEqual(sms.secretary_client, sentinel.secretary_client)

    def test_render_without_secretary_client(self):
        sms = Sms(recipient='recipient@example.com', body_template='body_template')
        with self.assertRaises(RuntimeError):
            sms.body

    def test_to_dict(self):
        sms = Sms.from_dict({
            'recipient': sentinel.recipient,
            'body_template': sentinel.body_template,
            'body_template_uuid': UUID(int=1),
            'context': {},
            'type': sentinel.type,
        })
        self.assertEqual(sms.to_dict(), {
            'recipient': sentinel.recipient,
            'body_template': sentinel.body_template,
            'body_template_uuid': UUID(int=1),
            'context': {},
            'type': sentinel.type,
        })


class SmsWithSecretaryClientTest(TestCase):
    """Unittests for Sms structure with a secretary client."""

    def setUp(self):
        self.secretary_client = Mock(autospec=SecretaryClient)
        self.sms = Sms(
            recipient='recipient@example.com',
            body_template='body_template',
            context=sentinel.context,
            type=sentinel.type,
            secretary_client=self.secretary_client,
        )

    def test_body(self):
        self.secretary_client.render.return_value = sentinel.body
        self.assertEqual(self.sms.body, sentinel.body)
        self.assertEqual(self.sms.body, sentinel.body)  # call body again to test the caching
        self.assertEqual(
            self.secretary_client.mock_calls,
            [call.render('body_template', context=sentinel.context)]
        )

    def test_to_dict(self):
        self.secretary_client.render.return_value = sentinel.body
        self.assertEqual(self.sms.to_dict(), {
            'recipient': 'recipient@example.com',
            'body_template': 'body_template',
            'body_template_uuid': None,
            'body': sentinel.body,
            'context': sentinel.context,
            'type': sentinel.type,
        })
